import PropTypes from 'prop-types';
import React, { createContext, useRef, useCallback, useEffect, useMemo, useContext, useState, useLayoutEffect } from 'react';

let StatusEnum;

(function (StatusEnum) {
  StatusEnum["ERROR"] = "error";
  StatusEnum["IDLE"] = "idle";
  StatusEnum["LOADING"] = "loading";
  StatusEnum["SUCCESS"] = "success";
})(StatusEnum || (StatusEnum = {}));

const KEY_IS_NOT_STRING_ERROR = 'Key should be a string';
const DEFAULT_MAX_CONCURRENT_REQUESTS = 20;

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

class DataLoader {
  constructor(args) {
    var _this = this;

    _defineProperty(this, "needPolling", true);

    _defineProperty(this, "isDataOutdated", false);

    _defineProperty(this, "canceled", false);

    _defineProperty(this, "errorListeners", []);

    _defineProperty(this, "successListeners", []);

    _defineProperty(this, "cancelListeners", []);

    _defineProperty(this, "observerListeners", []);

    _defineProperty(this, "destroyed", false);

    _defineProperty(this, "load", async function () {
      let force = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      if (force || _this.status !== StatusEnum.SUCCESS || _this.status === StatusEnum.SUCCESS && _this.isDataOutdated) {
        if (_this.timeout) {
          // Prevent multiple call at the same time
          clearTimeout(_this.timeout);
        }

        await _this.launch();
      }
    });

    _defineProperty(this, "launch", async () => {
      try {
        if (this.status !== StatusEnum.LOADING) {
          this.canceled = false;
          this.status = StatusEnum.LOADING;
          this.notifyChanges();
        }

        DataLoader.started += 1;
        const promise = this.method();
        this.cancelMethod = promise.cancel;
        const result = await promise.then(res => res);
        this.status = StatusEnum.SUCCESS;
        this.error = undefined;

        if (!this.canceled) {
          DataLoader.cachedData[this.key] = result;
          await Promise.all(this.successListeners.map(listener => listener === null || listener === void 0 ? void 0 : listener(result)));
          this.isDataOutdated = false;

          if (this.dataOutdatedTimeout) {
            clearTimeout(this.dataOutdatedTimeout);
            this.dataOutdatedTimeout = undefined;
          }

          if (this.maxDataLifetime) {
            this.dataOutdatedTimeout = setTimeout(() => {
              this.isDataOutdated = true;
              this.notifyChanges();
            }, this.maxDataLifetime);
          }
        }

        this.notifyChanges();
      } catch (err) {
        this.status = StatusEnum.ERROR;
        this.error = err;
        this.notifyChanges();

        if (!this.canceled) {
          await Promise.all(this.errorListeners.map(listener => listener === null || listener === void 0 ? void 0 : listener(err)));
        }
      }

      DataLoader.started -= 1;

      if (this.pollingInterval && !this.destroyed && (typeof this.needPolling === 'function' ? this.needPolling(DataLoader.cachedData[this.key]) : this.needPolling)) {
        this.timeout = setTimeout( // eslint-disable-next-line @typescript-eslint/no-misused-promises
        this.launch, this.pollingInterval);
      }
    });

    _defineProperty(this, "cancel", async () => {
      var _this$cancelMethod;

      this.canceled = true;
      (_this$cancelMethod = this.cancelMethod) === null || _this$cancelMethod === void 0 ? void 0 : _this$cancelMethod.call(this);
      await Promise.all(this.cancelListeners.map(listener => listener === null || listener === void 0 ? void 0 : listener()));
    });

    this.key = args.key;
    this.enabled = args.enabled ?? false;
    this.status = args.enabled ? StatusEnum.LOADING : StatusEnum.IDLE;
    this.method = args.method;
    this.pollingInterval = args === null || args === void 0 ? void 0 : args.pollingInterval;
    this.keepPreviousData = args === null || args === void 0 ? void 0 : args.keepPreviousData;
    this.maxDataLifetime = args.maxDataLifetime;
    this.needPolling = args.needPolling ?? true;

    if (args.enabled) {
      this.tryLaunch();
    } else {
      this.notifyChanges();
    }
  }

  tryLaunch() {
    if (DataLoader.started < DataLoader.maxConcurrent) {
      // Because we want to launch the request directly without waiting the return
      // eslint-disable-next-line no-void
      void this.load();
    } else {
      setTimeout(() => this.tryLaunch());
    }
  }

  getData() {
    return DataLoader.cachedData[this.key] ?? undefined;
  }

  notifyChanges() {
    this.observerListeners.forEach(observerListener => observerListener(this));
  }

  addOnSuccessListener(fn) {
    if (!this.successListeners.includes(fn)) {
      this.successListeners.push(fn);
    }
  }

  addOnErrorListener(fn) {
    if (!this.errorListeners.includes(fn)) {
      this.errorListeners.push(fn);
    }
  }

  addOnCancelListener(fn) {
    if (!this.cancelListeners.includes(fn)) {
      this.cancelListeners.push(fn);
    }
  }

  removeOnSuccessListener(fn) {
    const index = this.successListeners.indexOf(fn);

    if (index > -1) {
      this.successListeners.splice(index, 1);
    }
  }

  removeOnErrorListener(fn) {
    const index = this.errorListeners.indexOf(fn);

    if (index > -1) {
      this.errorListeners.splice(index, 1);
    }
  }

  removeOnCancelListener(fn) {
    const index = this.cancelListeners.indexOf(fn);

    if (index > -1) {
      this.cancelListeners.splice(index, 1);
    }
  }

  addObserver(fn) {
    this.observerListeners.push(fn);
  }

  removeObserver(fn) {
    const index = this.observerListeners.indexOf(fn);

    if (index > -1) {
      this.observerListeners.splice(index, 1);
    }
  }

  clearData() {
    DataLoader.cachedData[this.key] = undefined;
    this.notifyChanges();
  }

  async destroy() {
    var _this$cancel;

    DataLoader.cachedData[this.key] = undefined;
    await ((_this$cancel = this.cancel) === null || _this$cancel === void 0 ? void 0 : _this$cancel.call(this));

    if (this.timeout) {
      clearTimeout(this.timeout);
    }

    this.destroyed = true;
  }

  getObserversCount() {
    return this.observerListeners.length;
  }

  setPollingInterval(newPollingInterval) {
    this.pollingInterval = newPollingInterval;

    if (this.enabled) {
      this.tryLaunch();
    }
  }

  setNeedPolling(needPolling) {
    this.needPolling = needPolling;
  }

}

_defineProperty(DataLoader, "started", 0);

_defineProperty(DataLoader, "maxConcurrent", DEFAULT_MAX_CONCURRENT_REQUESTS);

_defineProperty(DataLoader, "cachedData", {});

// @ts-expect-error we force the context to undefined, should be corrected with default values
const DataLoaderContext = createContext(undefined);

const DataLoaderProvider = _ref => {
  let {
    children,
    cacheKeyPrefix,
    onError,
    maxConcurrentRequests
  } = _ref;
  const requestsRef = useRef({});
  const computeKey = useCallback(key => `${cacheKeyPrefix ? `${cacheKeyPrefix}-` : ''}${key}`, [cacheKeyPrefix]);
  const getRequest = useCallback(key => requestsRef.current[computeKey(key)], [computeKey]);
  const addRequest = useCallback((key, args) => {
    if (DataLoader.maxConcurrent !== maxConcurrentRequests) {
      DataLoader.maxConcurrent = maxConcurrentRequests;
    }

    if (key && typeof key === 'string') {
      const newRequest = new DataLoader({ ...args,
        key: computeKey(key)
      });
      requestsRef.current[newRequest.key] = newRequest;
      return newRequest;
    }

    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey, maxConcurrentRequests]);
  const getOrAddRequest = useCallback((key, args) => {
    const requestFound = getRequest(key);

    if (!requestFound) {
      return addRequest(key, args);
    }

    return requestFound;
  }, [addRequest, getRequest]);
  const clearCachedData = useCallback(key => {
    if (typeof key === 'string') {
      if (requestsRef.current[computeKey(key)]) {
        requestsRef.current[computeKey(key)].clearData();
      }
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey]);
  const clearAllCachedData = useCallback(() => {
    Object.values(requestsRef.current).forEach(request => {
      request.clearData();
    });
  }, []);
  const reload = useCallback(async key => {
    if (key && typeof key === 'string') {
      var _getRequest;

      await ((_getRequest = getRequest(key)) === null || _getRequest === void 0 ? void 0 : _getRequest.load(true));
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [getRequest]);
  const reloadAll = useCallback(async () => {
    await Promise.all(Object.values(requestsRef.current).map(request => request.load(true)));
  }, []);
  const getCachedData = useCallback(key => {
    if (key) {
      var _getRequest2;

      return (_getRequest2 = getRequest(key)) === null || _getRequest2 === void 0 ? void 0 : _getRequest2.getData();
    }

    return Object.values(requestsRef.current).reduce((acc, request) => ({ ...acc,
      [request.key]: request.getData()
    }), {});
  }, [getRequest]);
  const getReloads = useCallback(key => {
    if (key) {
      return getRequest(key) ? () => getRequest(key).load(true) : undefined;
    }

    return Object.entries(requestsRef.current).reduce((acc, _ref2) => {
      let [requestKey, {
        load
      }] = _ref2;
      return { ...acc,
        [requestKey]: () => load(true)
      };
    }, {});
  }, [getRequest]);
  useEffect(() => {
    const cleanRequest = () => {
      setTimeout(() => {
        Object.keys(requestsRef.current).forEach(key => {
          if (requestsRef.current[key].getObserversCount() === 0) {
            // Worst case there is a memleak
            // eslint-disable-next-line @typescript-eslint/no-floating-promises
            requestsRef.current[key].destroy();
            delete requestsRef.current[key];
          }
        });
        cleanRequest();
      }, 300);
    };

    cleanRequest();
  }, []);
  const value = useMemo(() => ({
    addRequest,
    cacheKeyPrefix,
    clearAllCachedData,
    clearCachedData,
    getCachedData,
    getOrAddRequest,
    getReloads,
    getRequest,
    onError,
    reload,
    reloadAll
  }), [addRequest, cacheKeyPrefix, clearAllCachedData, clearCachedData, getCachedData, getOrAddRequest, getRequest, getReloads, onError, reload, reloadAll]);
  return /*#__PURE__*/React.createElement(DataLoaderContext.Provider, {
    value: value
  }, children);
};

DataLoaderProvider.propTypes = {
  cacheKeyPrefix: PropTypes.string,
  children: PropTypes.node.isRequired,
  maxConcurrentRequests: PropTypes.number,
  onError: PropTypes.func
};
DataLoaderProvider.defaultProps = {
  cacheKeyPrefix: undefined,
  maxConcurrentRequests: DEFAULT_MAX_CONCURRENT_REQUESTS,
  onError: undefined
};
const useDataLoaderContext = () => useContext(DataLoaderContext);

/**
 * @param {string} key key to save the data fetched in a local cache
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */
const useDataLoader = function (fetchKey, method) {
  let {
    enabled = true,
    initialData,
    keepPreviousData = true,
    onError,
    onSuccess,
    pollingInterval,
    maxDataLifetime,
    needPolling
  } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  const isMountedRef = useRef(false);
  const isFetchingRef = useRef(false);
  const unsubscribeRequestRef = useRef();
  const previousDataRef = useRef();
  const {
    getOrAddRequest,
    onError: onErrorProvider
  } = useDataLoaderContext();
  const [, forceReload] = useState(0);
  const subscribeFn = useCallback(() => {
    forceReload(x => x + 1);
  }, []);
  const request = useMemo(() => {
    if (unsubscribeRequestRef.current) {
      unsubscribeRequestRef.current();
    }

    const newRequest = getOrAddRequest(fetchKey, {
      enabled,
      keepPreviousData,
      maxDataLifetime,
      method,
      needPolling,
      pollingInterval
    });

    unsubscribeRequestRef.current = () => newRequest.removeObserver(subscribeFn);

    newRequest.addObserver(subscribeFn);
    return newRequest;
  }, [enabled, fetchKey, getOrAddRequest, maxDataLifetime, method, needPolling, pollingInterval, keepPreviousData, subscribeFn]);
  useEffect(() => {
    if (request.method !== method) {
      request.method = method;
    }

    request.addOnErrorListener(onError ?? onErrorProvider);
    request.addOnSuccessListener(onSuccess);
    return () => {
      request.removeOnErrorListener(onError ?? onErrorProvider);
      request.removeOnSuccessListener(onSuccess);
    };
  }, [onSuccess, onError, onErrorProvider, method, request]);
  const cancelMethodRef = useRef(request === null || request === void 0 ? void 0 : request.cancel);
  const isLoading = useMemo(() => request.status === StatusEnum.LOADING, [request.status]);
  const isIdle = useMemo(() => request.status === StatusEnum.IDLE, [request.status]);
  const isSuccess = useMemo(() => request.status === StatusEnum.SUCCESS, [request.status]);
  const isError = useMemo(() => request.status === StatusEnum.ERROR, [request.status]);
  const isPolling = useMemo(() => !!(enabled && pollingInterval && (isSuccess || isLoading)), [isSuccess, isLoading, enabled, pollingInterval]);
  useEffect(() => {
    if (enabled && isIdle) {
      var _request$launch;

      // launch should never throw
      // eslint-disable-next-line @typescript-eslint/no-floating-promises
      (_request$launch = request.launch) === null || _request$launch === void 0 ? void 0 : _request$launch.call(request);
    }
  }, [request, enabled, isIdle]);
  useEffect(() => {
    if (pollingInterval !== request.pollingInterval) {
      request.setPollingInterval(pollingInterval);
    }
  }, [pollingInterval, request]);
  useEffect(() => {
    if (needPolling !== request.needPolling) {
      request.setNeedPolling(needPolling ?? true);
    }
  }, [needPolling, request]);
  useEffect(() => {
    isFetchingRef.current = isLoading || isPolling;
  }, [isLoading, isPolling]);
  useLayoutEffect(() => {
    cancelMethodRef.current = request === null || request === void 0 ? void 0 : request.cancel;
  }, [request === null || request === void 0 ? void 0 : request.cancel]);
  useEffect(() => {
    isMountedRef.current = true;
    return () => {
      var _unsubscribeRequestRe;

      isMountedRef.current = false;

      if (isFetchingRef.current && cancelMethodRef.current) {
        cancelMethodRef.current();
      }

      (_unsubscribeRequestRe = unsubscribeRequestRef.current) === null || _unsubscribeRequestRe === void 0 ? void 0 : _unsubscribeRequestRe.call(unsubscribeRequestRef);
    };
  }, []);
  useEffect(() => () => {
    if (request.getData() !== previousDataRef.current && keepPreviousData) {
      previousDataRef.current = request.getData();
    }
  });
  return {
    data: request.getData() || initialData,
    error: request === null || request === void 0 ? void 0 : request.error,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    previousData: previousDataRef.current,
    reload: () => request.load(true)
  };
};

export { DataLoaderProvider, useDataLoader, useDataLoaderContext };
